/* bits.c (c) alain.spineux@gmail.com
 *
 * handle array of bits
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "bits.h"

void mask_for_last(long long int size, bit_int *mask)
{
    *mask=0;
    size=size%(8*sizeof(bit_int));
    int off=0;

    if (size==0) *mask=~0;
    else
    {
        while (off<size)
        {
            bit_int v=BE_TO_BIT_INT(1<<(8*sizeof(bit_int)-1-off));
            *mask=*mask | v;
            off++;
        }
    }
}

/**
 * a <=> b
 *
static bit_int biconditional(bit_int a, bit_int b)
{
    return ((~a)|b) & (a|(~b));
}
 */

/**
 * a => b   also, is a included in b ?
 */
static bit_int implication(bit_int a, bit_int b)
{
    return (~a)|b;
}

unsigned char bit_count[256];
static int bit_counter_initialized=0;

void init_bit_count()
{
	int i;
	for (i=0; i<256; i++)
	{
		unsigned char c=0;
		if (i&0x01) c++;
		if (i&0x02) c++;
		if (i&0x04) c++;
		if (i&0x08) c++;
		if (i&0x10) c++;
		if (i&0x20) c++;
		if (i&0x40) c++;
		if (i&0x80) c++;
		bit_count[i]=c;

	}
	bit_counter_initialized=1;
}

/**
 * compare bits in a and b
 */
static void bit_cmp(bit_int a, bit_int b, long long int *onlya, long long int *onlyb, long long int *both)
{
	if (!bit_counter_initialized) init_bit_count();
	int i;
	unsigned char *pa=(unsigned char *)&a;
	unsigned char *pb=(unsigned char *)&b;
	for (i=0; i<sizeof(bit_int); i++, pa++, pb++)
	{
		*both+=bit_count[(*pa)&(*pb)];
		*onlya+=bit_count[(*pa)&~(*pb)];
		*onlyb+=bit_count[(*pb)&~(*pa)];
	}
}

int bit_array_init(struct bit_array *ba, long long int size, int pattern)
{
    ba->size=size;
    ba->isize=(size+(8*sizeof(bit_int)-1))/8/sizeof(bit_int);
    ba->array=malloc(ba->isize*sizeof(bit_int));
    if (!ba->array) return 1;
    ba->allocated=1;

    ba->index=0;
    mask_for_last(size, &(ba->mask_last));
    bit_array_reset(ba, pattern);
    return 0;
}

int bit_array_init2(struct bit_array *ba, long long int size, void *buffer)
{
    ba->size=size;
    ba->isize=(size+(8*sizeof(bit_int)-1))/8/sizeof(bit_int);
    ba->array=(bit_int *)buffer;
    ba->allocated=0;
    ba->index=0;
    mask_for_last(size, &(ba->mask_last));
    return 0;
}

void bit_array_release(struct bit_array *ba)
{
    if (ba->allocated) free(ba->array);
}

void bit_array_reset(struct bit_array *ba, int pattern)
{
    long long int i;
    bit_int value=0;
    if (pattern) value=~value;
    bit_int *p=ba->array;

    for (i=0; i<ba->isize; i++, p++) *p=value;
}

int bit_array_set(struct bit_array *ba, long long int bit_addr)
{
    if (bit_addr>=ba->size) return 0;
    int off=bit_addr%(8*sizeof(bit_int));

    bit_int *p=ba->array+(bit_addr/(8*sizeof(bit_int)));
    bit_int v=BIT_INT_TO_BE(1<<(8*sizeof(bit_int)-1-off));
    int res=*p & v;
    *p=*p | v;
    return res;
}

int bit_array_unset(struct bit_array *ba, long long int bit_addr)
{
    if (bit_addr>=ba->size) return 0;
    int off=bit_addr%(8*sizeof(bit_int));

    bit_int *p=ba->array+(bit_addr/(8*sizeof(bit_int)));
    bit_int v=BIT_INT_TO_BE(~(1<<(8*sizeof(bit_int)-1-off)));
    int res=*p & ~v;
    *p=*p & v;
    return res;
}

int bit_array_get(struct bit_array *ba, long long int bit_addr)
{
    if (bit_addr>=ba->size) return 0;
    int off=bit_addr%(8*sizeof(bit_int));

    bit_int *p=ba->array+(bit_addr/(8*sizeof(bit_int)));
    return (*p & BIT_INT_TO_BE(1<<(8*sizeof(bit_int)-1-off)))!=0;
}

void bit_array_print(struct bit_array *ba)
{
    long long int i;
    long long int k=ba->size;
    int j;

    printf("%lld %lld : ", ba->size, ba->isize);

    for (i=0; i<ba->isize; i++)
    {
        bit_int v=BE_TO_BIT_INT(ba->array[i]);
        for (j=0; j<8*sizeof(bit_int) && k>0; j++, k--)
        {
            if (v & BIT_INT_HIGHEST_BIT) printf("1");
            else printf("0");
            v=v<<1;
        }
    }
    printf("\n");
}

long long int bit_array_search_first_set(struct bit_array *ba, long long int from)
{
    long long int addr;
    long long int i=from/(8*sizeof(bit_int));
    int         off=from%(8*sizeof(bit_int));
    int           j;

    bit_int v;
    while (i*(8*sizeof(bit_int))+off<ba->size)
    {
        while (i*(8*sizeof(bit_int))<ba->size && ba->array[i]==0)
        {
            i++;
            off=0;
        }
        if (i*(8*sizeof(bit_int))>=ba->size) return -1;
        v=BE_TO_BIT_INT(ba->array[i])<<off;
        for (j=off; j<(8*sizeof(bit_int)); j++)
        {
            if (v & BIT_INT_HIGHEST_BIT) {
                addr=i*(8*sizeof(bit_int))+j;
                if (addr<ba->size) return addr;
                else return -1;
            }
            v=v<<1;
        }
        i++;
        off=0;
    }
    return -1;
}

long long int bit_array_search_first_unset(struct bit_array *ba, long long int from)
{
    long long int addr;
    long long int i=from/(8*sizeof(bit_int));
    int         off=from%(8*sizeof(bit_int));
    int           j;

    bit_int v;
    bit_int inv=~0;
    while (i*(8*sizeof(bit_int))+off<ba->size)
    {
        while (i*(8*sizeof(bit_int))<ba->size && ba->array[i]==inv)
        {
            i++;
            off=0;
        }
        if (i*(8*sizeof(bit_int))>=ba->size) return -1;
        v=BE_TO_BIT_INT(ba->array[i])<<off;
        for (j=off; j<(8*sizeof(bit_int)); j++)
        {
            if ((v & BIT_INT_HIGHEST_BIT)==0) {
                addr=i*(8*sizeof(bit_int))+j;
                if (addr<ba->size) return addr;
                else return -1;
            }
            v=v<<1;
        }
        i++;
        off=0;
    }
    return -1;
}

void bit_array_invers_into(struct bit_array *src, struct bit_array *dst)
{
    long long int i;
    bit_int *p, *q;

    for (i=0, p=src->array, q=dst->array; i<src->isize; i++, p++, q++) *q=~*p;
}

void bit_array_bwand(struct bit_array *src, struct bit_array *dst)
{
    long long int i;
    bit_int *p, *q;

    for (i=0, p=src->array, q=dst->array; i<src->isize; i++, p++, q++) (*q)&=*p;
}

void bit_array_bwor(struct bit_array *src, struct bit_array *dst)
{
    long long int i;
    bit_int *p, *q;

    for (i=0, p=src->array, q=dst->array; i<src->isize; i++, p++, q++) (*q)|=*p;
}

void bit_array_plus_diff(struct bit_array *a, struct bit_array *b, struct bit_array *c)
{ // a=a+(b-c)

    long long int i;
    bit_int *p, *q, *r;

    for (i=0, p=a->array, q=b->array, r=c->array; i<a->isize; i++, p++, q++, r++) (*p)|=*q & ~*r;
}


void bit_array_copy(struct bit_array *src, struct bit_array *dst)
{
    long long int i;
    bit_int *p, *q;

    for (i=0, p=src->array, q=dst->array; i<src->isize; i++, p++, q++) *q=*p;
}

/**
 * compare 2 bit array
 *
 * if (bit_array_cmp(a, b, &cmp) && cmp==0) then equal
 * if (bit_array_cmp(a, b, &cmp) && cmp<0) then a<b
 * if (bit_array_cmp(a, b, &cmp) && cmp>0) then a>b
 * if (!bit_array_cmp(a, b, &cmp)) then nothing to say on a and b
 *
 * @param ba1
 * @param ba2
 * @param cmp if one is included into the other, return -1, 1 or 0 if
 * respectively ba1<ba2, ba1>ba2, ba1==ba2 otherwise 0
 * @return 1 if one is included into the other, 0 otherwise
 */
int bit_array_cmp(struct bit_array *ba1, struct bit_array *ba2, int *cmp)
{
    long long int i;
    bit_int *p, *q;
    bit_int cmp1=~(bit_int)0;
    bit_int cmp2=~(bit_int)0;

    //printf("|| cmp1=0x%08x cmp2=0x%08x\n", cmp1, cmp2);
    // start at 1
    for (i=1, p=ba1->array, q=ba2->array; i<ba1->isize; i++, p++, q++)
    {
        cmp1&=implication(*p, *q);
        cmp2&=implication(*q, *p);
        //printf("   cmp1=0x%08x cmp2=0x%08x\n", cmp1, cmp2);
    }
    // use mask for the last one
    cmp1&=implication((*p)&ba1->mask_last, (*q)&ba2->mask_last);
    cmp2&=implication((*q)&ba2->mask_last, (*p)&ba1->mask_last);
    //printf("   cmp1=0x%08x cmp2=0x%08x\n", cmp1, cmp2);

    cmp1=(cmp1==~(bit_int)0);
    cmp2=(cmp2==~(bit_int)0);

    //printf("-> cmp1=%d cmp2=%d\n", cmp1, cmp2);

    *cmp=0;
    if (cmp1) (*cmp)--;
    if (cmp2) (*cmp)++;
    if (cmp1 || cmp2) return 1;
    return 0;
}


/**
 * compare 2 bit array
 *
 * @param ba1
 * @param ba2
 * @param only1 number of bits only in ba1
 * @param only2 number of bits only in ba2
 * @param both number of bits in both
 * @return 1 if ba1==ba2
 */
int bit_array_cmp_count(struct bit_array *ba1, struct bit_array *ba2, long long int *only1, long long int *only2, long long int *both)
{
    long long int i;
	*only1=*only2=*both=0LL;
	int equal=1;

    bit_int *p, *q;

    // start at 1
    for (i=1, p=ba1->array, q=ba2->array; i<ba1->isize; i++, p++, q++)
    {
    	bit_cmp(*p, *q, only1, only2, both);
    	if (*p!=*q) equal=0;
    }
    // use mask for the last one
    bit_cmp((*p)&ba1->mask_last, (*q)&ba2->mask_last, only1, only2, both);
    if (((*p)&ba1->mask_last)!=((*q)&ba2->mask_last)) equal=0;
    return equal;
}


long long int bit_array_alloc(struct bit_array *ba)
{
    long long int addr=bit_array_search_first_unset(ba, ba->index);
    if (addr==-1) addr=bit_array_search_first_unset(ba, 0);

    if (addr==-1) ba->index=0;
    else
    {
        if (addr<ba->size) ba->index=addr+1;
        else ba->index=0;
        bit_array_set(ba, addr);
    }
    return addr;
}

void bit_array_count(struct bit_array *ba, long long int *s, long long int *u)
{
    long long int i;
    int j, n;
    bit_int inv=~0;
    bit_int v;
    *s=*u=0;
    i=0;
    n=(sizeof(bit_int)*8);
    while (i*(sizeof(bit_int)*8)<ba->size)
    {
        v=BE_TO_BIT_INT(ba->array[i]);
        if ((i+1)*(sizeof(bit_int)*8)>ba->size) n=(sizeof(bit_int)*8)-((i+1)*(sizeof(bit_int)*8))%ba->size;
        if (v==0) (*u)+=n;
        else if (v==inv) (*s)+=n;
        else
        {
            for (j=0; j<n; j++)
            {
                if (v & BIT_INT_HIGHEST_BIT) (*s)++;
                else (*u)++;
                v=v<<1;
            }
        }
        i++;
    }
    // assert(*s+*u==ba->size);
}

int bit_array_save(struct bit_array *ba, char *filename, int sync)
{
    int fd=open(filename, O_RDWR|O_CREAT|O_TRUNC, 0600);
    if (fd==-1) return -errno;
    int size=ba->isize*sizeof(bit_int);
    int len=write(fd, ba->array, size);
    if (len==size && sync) fsync(fd);
    close(fd);

    if (len==-1) return -errno;
    if (len!=size) return 1;
    return 0;
}

int bit_array_load(struct bit_array *ba, char *filename)
{
    int fd=open(filename, O_RDONLY);
    if (fd==-1) return -errno;
    int size=ba->isize*sizeof(bit_int);
    int len=read(fd, ba->array, size);
    close(fd);

    if (len==-1) return -errno;
    if (len!=size) return 1;
    return 0;
}


static void test_bit_cmp(bit_int a, bit_int b)
{
	long long int first, second, both;
	first=second=both=0LL;
	bit_cmp(a, b, &first, &second, &both);
	printf("cmp(0x%llx, 0x%llx) -> ( %lld, %lld, %lld )\n", (long long int)a, (long long int)b, first, second, both);
}


int ba_main(int argc, char *argv[])
{
    struct bit_array ba, bb;
    int test_size=60;

	init_bit_count();

    if (1)
    {
    	printf("bit_count(0x00)=%d\n", (int)bit_count[0x00]);
		printf("bit_count(0x10)=%d\n", (int)bit_count[0x10]);
		printf("bit_count(0x11)=%d\n", (int)bit_count[0x11]);
		printf("bit_count(0xFF)=%d\n", (int)bit_count[0xFF]);

    	bit_int zero=0;
    	bit_int full=~(bit_int)0;
    	bit_int two=0x10;
    	bit_int forth=0x1010;

    	test_bit_cmp(zero, zero);
    	test_bit_cmp(zero, two);
    	test_bit_cmp(zero, forth);
    	test_bit_cmp(zero, full);

    	test_bit_cmp(two, zero);
    	test_bit_cmp(two, two);
    	test_bit_cmp(two, forth);
    	test_bit_cmp(two, full);

    	test_bit_cmp(forth, zero);
    	test_bit_cmp(forth, two);
    	test_bit_cmp(forth, forth);
    	test_bit_cmp(forth, full);

    	test_bit_cmp(full, zero);
    	test_bit_cmp(full, two);
    	test_bit_cmp(full, forth);
    	test_bit_cmp(full, full);

    }


    if (1)
    {
        int value[]={ 1, 58, 4, 55, 13, 59, 0, 57, -1};
        int i;

        bit_array_init(&ba, test_size, 0);
        for (i=0; i<33; i++)
        {
            bit_int mask=0;
            mask_for_last(i, &mask);
            if (i>0) bit_array_set(&ba, i-1);
            // printf("mask %3d 0x%08x 0x%08x 0x%08x\n", i, mask, ba.array[0], (mask & ba.array[0]));
            assert((mask&ba.array[0])==ba.array[0]);
        }

        printf("test cmp\n");
        bit_array_init(&ba, test_size, 0);
        bit_array_init(&bb, test_size, 0);
        // printf("mask 0x%08x\n", ba.mask_last);
        int cmp;

        assert(bit_array_cmp(&ba, &bb, &cmp) && cmp==0);
        assert(bit_array_cmp(&bb, &ba, &cmp) && cmp==0);

        for (i=0; value[i]!=-1; i++)
        {
            printf("%d %d\n", i, value[i]);
            bit_array_set(&ba, value[i]);
            assert(bit_array_cmp(&ba, &bb, &cmp) && cmp>0);
            assert(bit_array_cmp(&bb, &ba, &cmp) && cmp<0);

            bit_array_set(&bb, value[i]);
            assert(bit_array_cmp(&ba, &bb, &cmp) && cmp==0);
            assert(bit_array_cmp(&bb, &ba, &cmp) && cmp==0);
        }
        printf("test cmp finished: ok\n");
    }

    if (1)
    {

        bit_array_init(&ba, test_size, 0);
        bit_array_print(&ba);
        bit_array_reset(&ba, 1);
        bit_array_print(&ba);

        int i=-1;
        while (1)
        {
        	i=bit_array_search_first_set(&ba, i+1);
        	printf("%d\n", i);
        	if (i==-1) break;
        }

    }


    if (0)
    {
        bit_array_init(&ba, test_size, 0);
        bit_array_init(&bb, test_size, 0);

        bit_array_print(&ba);
        bit_array_set(&ba, 1);
        bit_array_set(&ba, 58);
        bit_array_set(&ba, 4);
        bit_array_set(&ba, 55);
        bit_array_print(&ba);
        printf("search first set from 0 : %lld\n", bit_array_search_first_set(&ba, 0));
        printf("search first set from 1 : %lld\n", bit_array_search_first_set(&ba, 1));
        printf("search first set from 2 : %lld\n", bit_array_search_first_set(&ba, 2));
        printf("search first set from 15 : %lld\n", bit_array_search_first_set(&ba, 15));
        printf("search first set from 55 : %lld\n", bit_array_search_first_set(&ba, 55));
        printf("search first set from 56 : %lld\n", bit_array_search_first_set(&ba, 56));
        printf("search first set from 59 : %lld\n", bit_array_search_first_set(&ba, 59));
        bit_array_invers_into(&ba, &bb);
        bit_array_print(&bb);
        printf("search first unset from 0 : %lld\n", bit_array_search_first_unset(&bb, 0));
        printf("search first unset from 1 : %lld\n", bit_array_search_first_unset(&bb, 1));
        printf("search first unset from 2 : %lld\n", bit_array_search_first_unset(&bb, 2));
        printf("search first unset from 15 : %lld\n", bit_array_search_first_unset(&bb, 15));
        printf("search first unset from 55 : %lld\n", bit_array_search_first_unset(&bb, 55));
        printf("search first unset from 56 : %lld\n", bit_array_search_first_unset(&bb, 56));
        printf("search first unset from 59 : %lld\n", bit_array_search_first_unset(&bb, 59));

        bit_array_unset(&ba, 1);
        bit_array_unset(&ba, 58);
        bit_array_print(&ba);

        int i;
        long long int s, u;
        unsigned char buf[1024];

        bit_array_save(&ba, "/tmp/bit_array", 1);
        for (i=0; i<ba.isize*sizeof(bit_int); i++) printf("%02X", (int)buf[i]);
        printf("\n");
        bit_array_save(&bb, "/tmp/bit_array", 1);
        bit_array_print(&bb);
        bit_array_count(&bb, &s, &u);
        printf("count %lld + %lld = %lld\n", s, u, s+u);
        for (i=0; i<test_size-2; i++) printf("%lld ", bit_array_alloc(&bb));
        bit_array_count(&bb, &s, &u);
        printf("\ncount %lld + %lld = %lld\n", s, u, s+u);

        bit_array_release(&ba);
        bit_array_release(&bb);
    }

    if (0)
    {
        struct bit_array big;
        long long int n=2097152;
        bit_array_init(&big, n, 0);
        long long int i;

        printf("init\n");
        clock_t start, end;

        start=clock();
        bit_array_set(&big, n/3);
        bit_array_set(&big, 2*n/3);
        long long int p1, p2;
        for (i=0; i<10000; i++)
        {
            p1=bit_array_search_first_set(&big, 0);
            p2=bit_array_search_first_set(&big, p1+1);
        }
        end=clock();
        printf("search first set from 0    : %lld\n", p1);
        printf("search first set from prev : %lld\n", p2);

        printf("%lld\n",(long long int)(end-start));
    }
    return 0;

}
